﻿using System;
using System.Runtime.InteropServices;

namespace SvdCli.Storage
{
    public class UnmanagedGigabyteRamDisk : IStorage
    {
        private readonly IntPtr[] _blocks;
        private bool _isInit;
        public string ProductId { get; } = "Svd RamDisk";
        public string ProductVersion { get; } = "1.0";
        public string TypeDescription { get; } = "RamDisk";
        public uint BlockSize { get; } = 512;
        public ulong BlockCount { get; }
        public uint MaxTransferLength { get; } = 64 * CapacityUnits.KiloByte;
        public bool SupportTrim { get; } = false;
        public bool WriteProtect { get; set; } = false;

        public void Init()
        {
            if (_isInit) return;

            try
            {
                for (var i = 0; i < _blocks.Length; i++)
                {
                    _blocks[i] = Marshal.AllocHGlobal(CapacityUnits.Gigabyte);
                }

                _isInit = true;
            }
            catch (OutOfMemoryException)
            {
                Exit();
                throw;
            }
        }

        public void Exit()
        {
            foreach (var block in _blocks)
            {
                if (block != IntPtr.Zero) Marshal.FreeHGlobal(block);
            }
            GC.Collect();
            GC.WaitForPendingFinalizers();
            _isInit = false;
        }

        public UnmanagedGigabyteRamDisk(int sizeInGigabyte)
        {
            _blocks = new IntPtr[sizeInGigabyte];
            var size = (ulong)sizeInGigabyte * CapacityUnits.Gigabyte;
            BlockCount = size / BlockSize;
        }

        private void ReadInternal(byte[] buffer, int offset, int count, int gigabyteIndex, int gigabyteOffset)
        {
            Marshal.Copy(_blocks[gigabyteIndex] + gigabyteOffset, buffer, offset, count);
        }

        private void WriteInternal(byte[] buffer, int offset, int count, int gigabyteIndex, int gigabyteOffset)
        {
            Marshal.Copy(buffer, offset, _blocks[gigabyteIndex] + gigabyteOffset, count);
        }

        public void Read(byte[] buffer, ulong blockIndex, uint blockCount)
        {
            var rawOffset = blockIndex * BlockSize;
            var rawSize = blockCount * BlockSize;

            for (var i = 0; i < rawSize;)
            {
                var gigabyteIndex = (int)(rawOffset / CapacityUnits.Gigabyte);
                var gigabyteOffset = (int)(rawOffset % CapacityUnits.Gigabyte);
                var bytesToRead = CapacityUnits.Gigabyte - gigabyteOffset;

                if (i + bytesToRead > rawSize)
                {
                    bytesToRead = (int)(rawSize - i);
                }

                ReadInternal(buffer, i, bytesToRead, gigabyteIndex, gigabyteOffset);

                i += bytesToRead;
                rawOffset += (ulong)bytesToRead;
            }
        }

        public void Write(byte[] buffer, ulong blockIndex, uint blockCount)
        {
            var rawOffset = blockIndex * BlockSize;

            for (var i = 0; i < buffer.Length;)
            {
                var gigabyteIndex = (int)(rawOffset / CapacityUnits.Gigabyte);
                var gigabyteOffset = (int)(rawOffset % CapacityUnits.Gigabyte);
                var bytesToWrite = CapacityUnits.Gigabyte - gigabyteOffset;

                if (i + bytesToWrite > buffer.Length)
                {
                    bytesToWrite = buffer.Length - i;
                }

                WriteInternal(buffer, i, bytesToWrite, gigabyteIndex, gigabyteOffset);

                i += bytesToWrite;
                rawOffset += (ulong)bytesToWrite;
            }
        }

        public void Flush()
        {
        }

        public void Trim(params TrimDescriptor[] descriptors)
        {
        }
    }
}
